from __future__ import annotations

import uuid

def new_stream_request_id(source_id: str) -> str:
    # request-id must stay opaque: generation no longer encodes actor/source
    # routing semantics in ID prefixes.
    _ = source_id
    return f"req-{uuid.uuid4().hex}"


__all__ = [
    "new_stream_request_id",
]
