from __future__ import annotations

from flownet.runtime.transport.tcp_transport import TcpPacketTransport


class NetworkInterface(TcpPacketTransport):
    """Backward-compatible alias for TCP packet transport."""


__all__ = ["NetworkInterface"]
